import pytest

from hera.data import HERAError
from hera.utils import (
    align_caret,
    format_int,
    from_u16,
    make_ansi,
    register_to_index,
    to_u16,
    to_u32,
)


def test_to_u16_with_max_negative():
    assert to_u16(-1) == 65535


def test_to_u16_with_min_negative():
    assert to_u16(-32768) == 32768


def test_to_u16_with_mid_sized_negative():
    assert to_u16(-1734) == 63802


def test_to_u16_with_another_negative():
    assert to_u16(-25043) == 40493


def test_to_u16_with_positive():
    assert to_u16(17) == 17


def test_to_u16_with_max_positive():
    assert to_u16(65535) == 65535


def test_to_u16_with_zero():
    assert to_u16(0) == 0


def test_to_u16_with_overflow():
    with pytest.raises(HERAError):
        to_u16(-32769)


def test_to_u16_with_another_overflow():
    with pytest.raises(HERAError):
        to_u16(-45000)


def test_to_u16_with_positive_overflow():
    with pytest.raises(HERAError):
        to_u16(65536)


def test_to_u16_with_another_positive_overflow():
    with pytest.raises(HERAError):
        to_u16(70000)


def test_from_u16_with_min_negative():
    assert from_u16(65535) == -1


def test_from_u16_with_max_negative():
    assert from_u16(32768) == -32768


def test_from_u16_with_mid_sized_negative():
    assert from_u16(63802) == -1734


def test_from_u16_with_another_negative():
    assert from_u16(40493) == -25043


def test_from_u16_with_positive():
    assert from_u16(17) == 17


def test_from_u16_with_zero():
    assert from_u16(0) == 0


def test_to_u32_with_small_positive():
    assert to_u32(7) == 7


def test_to_u32_with_large_positive():
    assert to_u32(100000) == 100000


def test_to_u32_with_max_positive():
    assert to_u32(4294967295) == 4294967295


def test_to_u32_with_small_negative():
    assert to_u32(-17) == 4294967279


def test_to_u32_with_max_negative():
    assert to_u32(-1) == 4294967295


def test_to_u32_with_min_negative():
    assert to_u32(-2147483648) == 2147483648


def test_to_u32_with_large_negative():
    assert to_u32(-128000) == 4294839296


def test_to_u32_with_zero():
    assert to_u32(0) == 0


def test_to_u32_with_overflow():
    with pytest.raises(HERAError):
        to_u32(-2147483649)


def test_to_u32_with_another_overflow():
    with pytest.raises(HERAError):
        to_u32(-3000000000)


def test_to_u32_with_positive_overflow():
    with pytest.raises(HERAError):
        to_u32(4294967296)


def test_to_u32_with_another_positive_overflow():
    with pytest.raises(HERAError):
        to_u32(5000000000)


def test_register_to_index_with_numbered_registers():
    for i in range(0, 16):
        assert register_to_index("R" + str(i)) == i
        assert register_to_index("r" + str(i)) == i


def test_register_to_index_with_named_registers():
    assert register_to_index("FP") == 14
    assert register_to_index("fp") == 14
    assert register_to_index("SP") == 15
    assert register_to_index("sp") == 15
    assert register_to_index("Rt") == 11
    assert register_to_index("rt") == 11
    assert register_to_index("PC_ret") == 13
    assert register_to_index("pc_ret") == 13
    assert register_to_index("FP_alt") == 12
    assert register_to_index("fp_alt") == 12


def test_register_to_index_with_invalid_register():
    with pytest.raises(HERAError) as e:
        register_to_index("R16")
    assert "R16" in str(e.value)


def test_align_caret():
    assert align_caret("\t\t  a", 5) == "\t\t  "


def test_make_ansi_red():
    assert make_ansi(31, 1) == "\033[31;1m"


def test_make_ansi_reset():
    assert make_ansi(0) == "\033[0m"


def test_format_int_with_small_positive():
    assert format_int(5) == "0x0005 = 5"


def test_format_int_with_ASCII_value():
    assert format_int(65) == "0x0041 = 65 = 'A'"


def test_format_int_with_large_positive():
    assert format_int(4000) == "0x0fa0 = 4000"


def test_format_int_with_negative():
    assert format_int(65535) == "0xffff = 65535 = -1"


def test_format_int_with_non_default_format():
    assert format_int(2, spec="bds") == "0b0000000000000010 = 2"



"""
The HERA debugger.

The shell interface to the debugger is defined in `shell.py`; this module defines the
basic debugging operations that the shell makes use of.

The code in this package makes an important distinction between "real ops" and "original
ops." Original ops are the HERA operations as they appear in the program that the user
wrote. Real ops are the original ops transformed by the preprocessor into something that
the virtual machine can actually run. For example, a single original SET op corresponds
to two real ops, SETLO and SETHI.

Internally, the debugger operates on real ops, but whenever it displays output to the
user, it must be in terms of original ops so that the user is not confused.

Author:  Ian Fisher (iafisher@fastmail.com)
Version: February 2019
"""
import copy

try:
    # On systems that don't have libreadline installed, this import results in an
    # ImportError, which can just be ignored (of course, readline key bindings won't
    # work, but there's nothing we can do about that).
    import readline  # noqa: F401
except ImportError:
    pass

from hera.data import Label, Program, Settings
from hera.op import AbstractOperation
from hera.vm import VirtualMachine


class Debugger:
    """
    A class for debugging. External users should generally use the `debug` function
    in `shell.py` instead of instantiating this class directly.
    """

    def __init__(self, program: Program, settings: Settings) -> None:
        self.settings = settings
        self.program = program
        self.symbol_table = program.symbol_table
        # A map from instruction numbers (i.e., possible values of the program counter)
        # to human-readable line numbers.
        self.breakpoints = {}  # type: Dict[int, str]
        self.vm = VirtualMachine()
        # How many CALLs without RETURNs?
        self.calls = 0
        # Back-up of the debugger's state, to implement the "undo" command. Implicitly
        # defines a linked list that traverses the debugger's entire history. Set to
        # None when no operations have been performed.
        self.old = None  # type: Optional[Debugger]

        # In the future, step-by-step execution of data operations may be supported, but
        # for now they're just executed before interactive debugging starts.
        for data_op in self.program.data:
            data_op.execute(self.vm)

    def save(self) -> None:
        self.old = copy.copy(self)
        self.old.symbol_table = self.program.symbol_table.copy()
        self.old.breakpoints = self.breakpoints.copy()
        self.old.vm = self.vm.copy()

    def get_breakpoints(self) -> "Dict[int, str]":
        """Get the breakpoints dictionary."""
        return self.breakpoints

    def set_breakpoint(self, b: int) -> None:
        """Set a breakpoint at the given instruction number (not line number)."""
        self.breakpoints[b] = self.instruction_number_to_location(b)

    def at_breakpoint(self) -> bool:
        """Return True if the debugger is currently at a breakpoint."""
        return not self.finished() and self.vm.pc in self.breakpoints

    def next(self, *, step: bool) -> None:
        """
        Advance the debugger by one original instruction.

        When `step` is True and the current operation is a CALL, the entire function
        call up to the matching RETURN is executed. When `step` is False, just the CALL
        operation is executed, and the debugger is left on the first operation of the
        body of the function.
        """
        if self.finished():
            return

        if not step and self.program.code[self.vm.pc].original.name == "CALL":
            calls = self.calls
            # step=True prevents infinite regress.
            self.next(step=True)
            while (
                not self.finished() and not self.at_breakpoint() and self.calls > calls
            ):
                self.next(step=True)
        else:
            for real_op in self.real_ops():
                if real_op.name == "CALL":
                    self.calls += 1
                elif real_op.name == "RETURN":
                    self.calls -= 1

                real_op.execute(self.vm)

    def reset(self) -> None:
        """Reset the internal state of the debugger."""
        self.vm.reset()

    def op(self, index=None) -> AbstractOperation:
        """
        Return the original operation at the given index, which defaults to the current
        program counter.
        """
        if index is None:
            index = self.vm.pc
        return self.program.code[index].original

    def real_ops(self) -> "List[AbstractOperation]":
        """
        Return the real operations that correspond to the current original operation.
        """
        original = self.op()
        end = self.vm.pc
        while end < len(self.program.code) and self.op(end) == original:
            end += 1

        return self.program.code[self.vm.pc : end]

    def location_to_instruction_number(self, b: str) -> int:
        """Resolve a user-supplied location string into an instruction number."""
        if b == ".":
            return self.vm.pc

        if ":" in b:
            path, lineno = b.split(":", maxsplit=1)
        else:
            path = self.op().loc.path
            lineno = b

        try:
            lineno_as_int = int(lineno)
        except ValueError:
            try:
                opno = self.program.symbol_table[b]
                assert isinstance(opno, Label)
                return opno
            except (KeyError, AssertionError):
                raise ValueError("could not locate label `{}`.".format(b)) from None
        else:
            for pc, op in enumerate(self.program.code):
                if op.loc.path == path and op.loc.line == lineno_as_int:
                    return pc

            raise ValueError("could not find corresponding line.")

    def instruction_number_to_location(self, b: int, *, append_label=True) -> str:
        """
        Turn an instruction number into a human-readable location string with the file
        path and line number.
        """
        op = self.program.code[b].original or self.program.code[b]
        path = "<stdin>" if op.loc.path == "-" else op.loc.path
        loc = path + ":" + str(op.loc.line)

        if append_label:
            # Look for a label corresponding to the breakpoint.
            label = self.find_label(b)
            if label is not None:
                return "{} ({})".format(loc, label)

        return loc

    def find_label(self, ino: int) -> "Optional[str]":
        """Find a label, if one exists, corresponding to the instruction number."""
        for symbol, value in self.program.symbol_table.items():
            if value == ino and isinstance(value, Label):
                return symbol
        return None

    def finished(self) -> bool:
        """Return True if the debugger has finished executing the program."""
        return self.vm.halted or self.vm.pc >= len(self.program.code)

    def empty(self) -> bool:
        """Return True if the debugged program is empty."""
        return len(self.program.code) == 0


def reverse_lookup_label(symbol_table: "Dict[str, int]", value: int) -> "Optional[str]":
    """
    Return the name of the label that maps to `value`, or None if no such label is
    found. Constants and data labels are ignored.
    """
    for k, v in symbol_table.items():
        if value == v and isinstance(v, Label):
            return k
    return None



import pytest
from .utils import helper

from hera.utils import to_u16
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_ADD_flags_and_pc(vm):
    vm.registers[2] = to_u16(-1)
    vm.registers[3] = 1

    helper(vm, "ADD(R1, R2, R3)")

    assert vm.pc == 1
    assert vm.registers[1] == 0
    assert not vm.flag_sign
    assert vm.flag_zero


def test_ADD_small_numbers(vm):
    vm.registers[2] = 20
    vm.registers[3] = 22

    helper(vm, "ADD(R1, R2, R3)")

    assert vm.registers[1] == 42
    assert not vm.flag_overflow
    assert not vm.flag_carry


def test_ADD_resulting_in_negative(vm):
    vm.registers[2] = to_u16(-14)
    vm.registers[3] = 8

    helper(vm, "ADD(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-6)
    assert vm.flag_sign
    assert not vm.flag_zero


def test_ADD_resulting_in_zero(vm):
    vm.registers[2] = to_u16(-4)
    vm.registers[3] = 4

    helper(vm, "ADD(R1, R2, R3)")

    assert vm.registers[1] == 0
    assert not vm.flag_sign
    assert vm.flag_zero


def test_ADD_resulting_in_overflow(vm):
    vm.registers[2] = 32767
    vm.registers[3] = 1

    helper(vm, "ADD(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-32768)
    assert vm.flag_overflow
    assert not vm.flag_carry


def test_ADD_resulting_in_big_overflow(vm):
    vm.registers[2] = 32767
    vm.registers[3] = 32767

    helper(vm, "ADD(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-2)
    assert vm.flag_overflow
    assert not vm.flag_carry


def test_ADD_resulting_in_negative_overflow(vm):
    vm.registers[2] = to_u16(-32768)
    vm.registers[3] = to_u16(-32768)

    helper(vm, "ADD(R1, R2, R3)")

    assert vm.registers[1] == 0
    assert vm.flag_overflow
    assert vm.flag_carry


def test_ADD_with_carry_flag_set(vm):
    vm.flag_carry = True
    vm.registers[2] = 5
    vm.registers[3] = 3

    helper(vm, "ADD(R1, R2, R3)")

    assert vm.registers[1] == 9
    assert not vm.flag_carry


def test_ADD_with_carry_and_block_flags_set(vm):
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.registers[2] = 5
    vm.registers[3] = 3

    helper(vm, "ADD(R1, R2, R3)")

    assert vm.registers[1] == 8
    assert not vm.flag_carry
    assert vm.flag_carry_block


def test_ADD_with_overflow_from_carry(vm):
    vm.flag_carry = True
    vm.registers[2] = 32760
    vm.registers[3] = 7

    helper(vm, "ADD(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-32768)
    assert vm.flag_overflow
    assert not vm.flag_carry


def test_SUB_flags_and_pc(vm):
    vm.flag_carry_block = True
    vm.registers[2] = 40
    vm.registers[3] = 100

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.pc == 1
    assert vm.registers[1] == to_u16(-60)
    assert vm.flag_sign
    assert not vm.flag_zero


def test_SUB_small_numbers(vm):
    vm.flag_carry_block = True
    vm.registers[2] = 64
    vm.registers[3] = 22

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == 42
    assert not vm.flag_overflow
    assert vm.flag_carry


def test_SUB_resulting_in_negative(vm):
    vm.flag_carry_block = True
    vm.registers[2] = to_u16(-64)
    vm.registers[3] = 22

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-86)
    assert vm.flag_sign
    assert not vm.flag_zero


def test_SUB_resulting_in_zero(vm):
    vm.flag_carry_block = True
    vm.registers[2] = to_u16(-37)
    vm.registers[3] = to_u16(-37)

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == 0
    assert not vm.flag_sign
    assert vm.flag_zero


def test_SUB_with_two_negatives(vm):
    vm.flag_carry_block = True
    vm.registers[2] = to_u16(-20)
    vm.registers[3] = to_u16(-40)

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == 20
    assert vm.flag_carry
    assert not vm.flag_overflow


def test_SUB_resulting_in_minimum_negative_overflow(vm):
    vm.flag_carry_block = True
    vm.registers[2] = to_u16(-32768)
    vm.registers[3] = 1

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == 32767
    assert vm.flag_carry
    assert vm.flag_overflow


def test_SUB_resulting_in_big_negative_overflow(vm):
    vm.flag_carry_block = True
    vm.registers[2] = to_u16(-32000)
    vm.registers[3] = 32000

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == 1536
    assert vm.flag_carry
    assert vm.flag_overflow


def test_SUB_resulting_in_maximum_negative_overflow(vm):
    vm.flag_carry_block = True
    vm.registers[2] = to_u16(-32768)
    vm.registers[3] = 32767

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == 1
    assert vm.flag_carry
    assert vm.flag_overflow


def test_SUB_resulting_in_minimum_positive_overflow(vm):
    vm.flag_carry_block = True
    vm.registers[2] = 32767
    vm.registers[3] = to_u16(-1)

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-32768)
    assert not vm.flag_carry
    assert vm.flag_overflow


def test_SUB_resulting_in_big_positive_overflow(vm):
    vm.flag_carry_block = True
    vm.registers[2] = 27500
    vm.registers[3] = to_u16(-7040)

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-30996)
    assert not vm.flag_carry
    assert vm.flag_overflow


def test_SUB_resulting_in_maximum_positive_overflow(vm):
    vm.flag_carry_block = True
    vm.registers[2] = 32767
    vm.registers[3] = to_u16(-32768)

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-1)
    assert not vm.flag_carry
    assert vm.flag_overflow


def test_SUB_with_implicit_borrow(vm):
    vm.registers[2] = 17
    vm.registers[3] = 5

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == 11


def test_SUB_with_no_carry_block_and_no_borrow(vm):
    vm.flag_carry = True
    vm.registers[2] = to_u16(-64)
    vm.registers[3] = 22

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-86)
    assert vm.flag_carry


def test_SUB_overflow_from_borrow(vm):
    vm.registers[2] = to_u16(-32767)
    vm.registers[3] = 1

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == 32767
    assert vm.flag_carry
    assert vm.flag_overflow


def test_SUB_overflow_takes_borrow_into_account(vm):
    vm.registers[2] = 10
    vm.registers[3] = 11

    helper(vm, "SUB(R1, R2, R3)")

    assert not vm.flag_overflow


def test_SUB_sets_carry_for_equal_operands(vm):
    vm.flag_carry = True
    vm.registers[2] = 12
    vm.registers[3] = 12

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == 0
    assert vm.flag_carry


def test_SUB_does_not_set_carry_for_equal_operands_with_no_incoming_carry(vm):
    vm.registers[2] = 12
    vm.registers[3] = 12

    helper(vm, "SUB(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-1)
    assert not vm.flag_carry


def test_MUL_with_small_positives(vm):
    vm.registers[2] = 4
    vm.registers[3] = 2

    helper(vm, "MUL(R1, R2, R3)")

    assert vm.registers[1] == 8


def test_MUL_with_large_positives(vm):
    vm.registers[2] = 4500
    vm.registers[3] = 3

    helper(vm, "MUL(R1, R2, R3)")

    assert vm.registers[1] == 13500


def test_MUL_with_small_negatives(vm):
    vm.registers[2] = to_u16(-5)
    vm.registers[3] = 3

    helper(vm, "MUL(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-15)


def test_MUL_with_large_negatives(vm):
    vm.registers[2] = 7
    vm.registers[3] = to_u16(-3100)

    helper(vm, "MUL(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-21700)


def test_MUL_resulting_in_signed_positive_overflow(vm):
    vm.registers[2] = 17000
    vm.registers[3] = 3

    helper(vm, "MUL(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-14536)
    assert not vm.flag_carry
    assert vm.flag_overflow


def test_MUL_resulting_in_unsigned_positive_overflow(vm):
    vm.registers[2] = 128
    vm.registers[3] = 749

    helper(vm, "MUL(R1, R2, R3)")

    assert vm.registers[1] == 30336
    assert vm.flag_carry
    assert vm.flag_overflow


def test_MUL_resulting_in_signed_negative_overflow(vm):
    vm.registers[2] = to_u16(-400)
    vm.registers[3] = to_u16(-200)

    helper(vm, "MUL(R1, R2, R3)")

    assert vm.registers[1] == 14464
    assert vm.flag_carry
    assert vm.flag_overflow


def test_MUL_ignores_carry_when_blocked(vm):
    vm.flag_carry_block = True
    vm.flag_carry = True
    vm.registers[2] = 4
    vm.registers[3] = 12

    helper(vm, "MUL(R1, R2, R3)")

    assert vm.registers[1] == 48
    assert not vm.flag_carry


def test_MUL_ignores_carry_when_not_blocked(vm):
    vm.flag_carry = True
    vm.registers[2] = 4
    vm.registers[3] = 12

    helper(vm, "MUL(R1, R2, R3)")

    assert vm.registers[1] == 48
    assert not vm.flag_carry


def test_MUL_produces_high_bits_when_sign_flag_is_on(vm):
    vm.flag_sign = True
    vm.registers[2] = 20000
    vm.registers[3] = 200

    helper(vm, "MUL(R1, R2, R3)")

    assert vm.registers[1] == 0b111101
    # TODO: Find out how the carry and overflow flags should be set.


def test_MUL_with_sign_flag_and_negative_result(vm):
    vm.flag_sign = True
    vm.registers[2] = 20000
    vm.registers[3] = to_u16(-200)

    helper(vm, "MUL(R1, R2, R3)")

    assert vm.registers[1] == 0b1111111111000010


def test_MUL_ignores_sign_flag_when_carry_is_blocked(vm):
    vm.flag_sign = True
    vm.flag_carry_block = True
    vm.registers[2] = 20000
    vm.registers[3] = 200

    helper(vm, "MUL(R1, R2, R3)")

    assert vm.registers[1] == 2304
    assert vm.flag_carry


def test_AND_same_numbers(vm):
    vm.registers[2] = 27
    vm.registers[3] = 27

    helper(vm, "AND(R1, R2, R3)")

    assert vm.registers[1] == 27


def test_AND_different_numbers(vm):
    vm.registers[2] = 0b011
    vm.registers[3] = 0b110

    helper(vm, "AND(R1, R2, R3)")

    assert vm.registers[1] == 0b010


def test_AND_big_numbers(vm):
    vm.registers[2] = 62434
    vm.registers[3] = 17589

    helper(vm, "AND(R1, R2, R3)")

    assert vm.registers[1] == 16544


def test_AND_does_not_set_other_flags(vm):
    vm.registers[2] = to_u16(-1)
    vm.registers[3] = to_u16(-1)

    helper(vm, "AND(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-1)
    assert not vm.flag_carry
    assert not vm.flag_overflow


def test_AND_does_not_clear_other_flags(vm):
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.registers[2] = to_u16(-1)
    vm.registers[3] = to_u16(-1)

    helper(vm, "AND(R1, R2, R3)")

    assert vm.flag_carry
    assert vm.flag_overflow


def test_OR_same_numbers(vm):
    vm.registers[2] = 27
    vm.registers[3] = 27

    helper(vm, "OR(R1, R2, R3)")

    assert vm.registers[1] == 27


def test_OR_different_numbers(vm):
    vm.registers[2] = 0b011
    vm.registers[3] = 0b110

    helper(vm, "OR(R1, R2, R3)")

    assert vm.registers[1] == 0b111


def test_OR_big_numbers(vm):
    vm.registers[2] = 8199
    vm.registers[3] = 762

    helper(vm, "OR(R1, R2, R3)")

    assert vm.registers[1] == 8959


def test_OR_does_not_set_other_flags(vm):
    vm.registers[2] = to_u16(-1)
    vm.registers[3] = to_u16(-1)

    helper(vm, "OR(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-1)
    assert not vm.flag_carry
    assert not vm.flag_overflow


def test_OR_does_not_clear_other_flags(vm):
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.registers[2] = to_u16(-1)
    vm.registers[3] = to_u16(-1)

    helper(vm, "OR(R1, R2, R3)")

    assert vm.flag_carry
    assert vm.flag_overflow


def test_XOR_same_numbers(vm):
    vm.registers[2] = 27
    vm.registers[3] = 27

    helper(vm, "XOR(R1, R2, R3)")

    assert vm.registers[1] == 0


def test_XOR_different_numbers(vm):
    vm.registers[2] = 0b011
    vm.registers[3] = 0b110

    helper(vm, "XOR(R1, R2, R3)")

    assert vm.registers[1] == 0b101


def test_XOR_big_numbers(vm):
    vm.registers[2] = 8199
    vm.registers[3] = 762

    helper(vm, "XOR(R1, R2, R3)")

    assert vm.registers[1] == 8957


def test_XOR_does_not_set_other_flags(vm):
    vm.registers[2] = 0
    vm.registers[3] = to_u16(-37)

    helper(vm, "XOR(R1, R2, R3)")

    assert vm.registers[1] == to_u16(-37)
    assert not vm.flag_carry
    assert not vm.flag_overflow


def test_XOR_does_not_clear_other_flags(vm):
    vm.flag_carry = True
    vm.flag_overflow = True
    vm.registers[2] = 0
    vm.registers[3] = to_u16(-37)

    helper(vm, "XOR(R1, R2, R3)")

    assert vm.flag_carry
    assert vm.flag_overflow



import pytest

from .utils import execute_program_helper


def test_parse_error_for_unclosed_string_literal(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper('LP_STRING("hello)')

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: unclosed string literal, line 1 col 11 of <stdin>

  LP_STRING("hello)
            ^

"""
    )


def test_parse_error_for_unclosed_string_literal_ending_with_backslash(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper('LP_STRING("devilish\\')

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: unclosed string literal, line 1 col 11 of <stdin>

  LP_STRING("devilish\\
            ^

"""
    )


def test_parse_error_for_unclosed_angle_include(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("#include <data.h")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: unclosed bracketed expression, line 1 col 11 of <stdin>

  #include <data.h
            ^

"""
    )


def test_parse_error_for_cpp_boilerplate_missing_left_parenthesis(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("void HERA_main) { }")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: expected left parenthesis, line 1 col 15 of <stdin>

  void HERA_main) { }
                ^

"""
    )


def test_parse_error_for_cpp_boilerplate_missing_right_parenthesis(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("void HERA_main( { }")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: expected right parenthesis, line 1 col 17 of <stdin>

  void HERA_main( { }
                  ^

"""
    )


def test_parse_error_for_cpp_boilerplate_missing_left_curly_brace(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("void HERA_main()  }")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: expected left curly brace, line 1 col 19 of <stdin>

  void HERA_main()  }
                    ^

"""
    )


def test_parse_error_for_error_in_cpp_boilerplate(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("void HERA_main) { SET(R1 1) }")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: expected left parenthesis, line 1 col 15 of <stdin>

  void HERA_main) { SET(R1 1) }
                ^

Error: expected comma or right parenthesis, line 1 col 26 of <stdin>

  void HERA_main) { SET(R1 1) }
                           ^

"""
    )


def test_parse_error_for_two_symbols_in_a_row(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("SET LO(R1, 1)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: expected left parenthesis, line 1 col 5 of <stdin>

  SET LO(R1, 1)
      ^

Error: unknown instruction `LO`, line 1 col 5 of <stdin>

  SET LO(R1, 1)
      ^

"""
    )


def test_parse_error_for_unclosed_arglist(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("SET(R1, 4\n")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: expected comma or right parenthesis, line 2 col 1 of <stdin>

  
  ^

"""
    )


def test_parse_error_for_trailing_comma_in_arglist(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("SET(R1, 4,)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: expected value, line 1 col 11 of <stdin>

  SET(R1, 4,)
            ^

"""
    )


def test_parse_error_for_arglist_is_just_a_comma(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("SET(,)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: expected value, line 1 col 5 of <stdin>

  SET(,)
      ^

Error: expected value, line 1 col 6 of <stdin>

  SET(,)
       ^

"""
    )


def test_parse_error_for_include_following_by_invalid_token(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("#include HERA")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: expected quote or angle-bracket delimited string, line 1 col 10 of <stdin>

  #include HERA
           ^

"""
    )


def test_parse_error_for_arithmetic_expression(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("SET(R1, 4*4)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: expected comma or right parenthesis, line 1 col 10 of <stdin>

  SET(R1, 4*4)
           ^

"""
    )


def test_parse_error_for_invalid_integer_literal(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("SET(R1, 0xg)")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: invalid integer literal, line 1 col 9 of <stdin>

  SET(R1, 0xg)
          ^

"""
    )


def test_parse_error_for_non_ASCII_byte(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper('LP_STRING("привет")')

    captured = capsys.readouterr().err
    assert captured == "\nError: non-ASCII byte in file.\n"


def test_parse_error_for_non_ASCII_byte_in_file(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper('#include "test/assets/error/non_ascii_byte.hera"')

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: non-ASCII byte in file "test/assets/error/non_ascii_byte.hera", line 1 col 10 of <stdin>

  #include "test/assets/error/non_ascii_byte.hera"
           ^

"""
    )


def test_parse_error_for_over_long_character_literal(capsys):
    with pytest.raises(SystemExit):
        execute_program_helper("SET(R1, 'abc')")

    captured = capsys.readouterr().err
    assert (
        captured
        == """\

Error: over-long character literal, line 1 col 9 of <stdin>

  SET(R1, 'abc')
          ^

"""
    )



import pytest
from .utils import helper

from hera.utils import to_u16
from hera.vm import VirtualMachine


@pytest.fixture
def vm():
    return VirtualMachine()


def test_LSL_with_small_positive(vm):
    vm.registers[2] = 7

    helper(vm, "LSL(R1, R2)")

    assert vm.registers[1] == 14


def test_LSL_with_large_positive(vm):
    vm.registers[2] = 15000

    helper(vm, "LSL(R1, R2)")

    assert vm.registers[1] == 30000


def test_LSL_with_positive_overflow(vm):
    vm.registers[2] = 17000

    helper(vm, "LSL(R1, R2)")

    assert vm.registers[1] == 34000


def test_LSL_with_small_negative(vm):
    vm.registers[2] = to_u16(-7)

    helper(vm, "LSL(R1, R2)")

    assert vm.registers[1] == to_u16(-14)
    assert vm.flag_sign
    assert not vm.flag_zero


def test_LSL_with_large_negative(vm):
    vm.registers[2] = to_u16(-8400)

    helper(vm, "LSL(R1, R2)")

    assert vm.registers[1] == to_u16(-16800)


def test_LSL_with_negative_overflow(vm):
    vm.registers[2] = to_u16(-20000)

    helper(vm, "LSL(R1, R2)")

    assert vm.registers[1] == 25536
    assert vm.flag_carry
    assert not vm.flag_overflow


def test_LSL_shifts_out_carry_when_blocked(vm):
    vm.flag_carry_block = True
    vm.registers[2] = to_u16(-20000)

    helper(vm, "LSL(R1, R2)")

    assert vm.flag_carry


def test_LSL_shifts_in_carry(vm):
    vm.flag_carry = True
    vm.registers[2] = 7

    helper(vm, "LSL(R1, R2)")

    assert vm.registers[1] == 15
    assert not vm.flag_carry


def test_LSL_ignores_carry_when_blocked(vm):
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.registers[2] = 7

    helper(vm, "LSL(R1, R2)")

    assert vm.registers[1] == 14
    assert not vm.flag_carry


def test_LSL_resets_carry(vm):
    vm.flag_carry = True

    helper(vm, "LSL(R1, R2)")

    assert not vm.flag_carry


def test_LSL_ignores_overflow_flag(vm):
    vm.flag_overflow = True

    helper(vm, "LSL(R1, R2)")

    assert vm.flag_overflow


def test_LSR_with_small_positive(vm):
    vm.registers[2] = 7

    helper(vm, "LSR(R1, R2)")

    assert vm.registers[1] == 3


def test_LSR_with_large_positive(vm):
    vm.registers[2] = 15000

    helper(vm, "LSR(R1, R2)")

    assert vm.registers[1] == 7500


def test_LSR_with_small_negative(vm):
    vm.registers[2] = to_u16(-7)

    helper(vm, "LSR(R1, R2)")

    assert vm.registers[1] == 32764


def test_LSR_with_large_negative(vm):
    vm.registers[2] = to_u16(-8400)

    helper(vm, "LSR(R1, R2)")

    assert vm.registers[1] == 28568


def test_LSR_with_another_large_negative(vm):
    vm.registers[2] = to_u16(-20000)

    helper(vm, "LSR(R1, R2)")

    assert vm.registers[1] == 22768


def test_LSR_shifts_out_carry_when_blocked(vm):
    vm.flag_carry_block = True
    vm.registers[2] = 3

    helper(vm, "LSR(R1, R2)")

    assert vm.registers[1] == 1
    assert vm.flag_carry


def test_LSR_shifts_in_carry(vm):
    vm.flag_carry = True
    vm.registers[2] = 6

    helper(vm, "LSR(R1, R2)")

    assert vm.registers[1] == to_u16(-32765)
    assert not vm.flag_carry


def test_LSR_ignores_carry_when_blocked(vm):
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.registers[2] = 6

    helper(vm, "LSR(R1, R2)")

    assert vm.registers[1] == 3
    assert not vm.flag_carry


def test_LSR_resets_carry(vm):
    vm.flag_carry = True

    helper(vm, "LSR(R1, R2)")

    assert not vm.flag_carry


def test_LSR_ignores_overflow_flag(vm):
    vm.flag_overflow = True

    helper(vm, "LSR(R1, R2)")

    assert vm.flag_overflow


def test_LSL8_with_small_positive(vm):
    vm.registers[2] = 51

    helper(vm, "LSL8(R1, R2)")

    assert vm.registers[1] == 13056


def test_LSL8_with_large_positive(vm):
    vm.registers[2] = 17000

    helper(vm, "LSL8(R1, R2)")

    assert vm.registers[1] == 26624


def test_LSL8_with_small_negative(vm):
    vm.registers[2] = to_u16(-4)

    helper(vm, "LSL8(R1, R2)")

    assert vm.registers[1] == to_u16(-1024)


def test_LSL8_with_large_negative(vm):
    vm.registers[2] = to_u16(-31781)

    helper(vm, "LSL8(R1, R2)")

    assert vm.registers[1] == to_u16(-9472)


def test_LSL8_ignores_incoming_carry(vm):
    vm.flag_carry = True
    vm.registers[2] = 5

    helper(vm, "LSL8(R1, R2)")

    assert vm.registers[1] == 1280
    assert vm.flag_carry


def test_LSL8_does_not_set_carry_or_overflow(vm):
    vm.registers[2] = to_u16(-1)

    helper(vm, "LSL8(R1, R2)")

    assert vm.registers[1] == to_u16(-256)
    assert not vm.flag_carry
    assert not vm.flag_overflow


def test_LSR8_with_small_positive(vm):
    vm.registers[2] = 51

    helper(vm, "LSR8(R1, R2)")

    assert vm.registers[1] == 0
    assert not vm.flag_sign
    assert vm.flag_zero


def test_LSR8_with_large_positive(vm):
    vm.registers[2] = 17000

    helper(vm, "LSR8(R1, R2)")

    assert vm.registers[1] == 66


def test_LSR8_with_small_negative(vm):
    vm.registers[2] = to_u16(-4)

    helper(vm, "LSR8(R1, R2)")

    assert vm.registers[1] == 255


def test_LSR8_with_large_negative(vm):
    vm.registers[2] = to_u16(-31781)

    helper(vm, "LSR8(R1, R2)")

    assert vm.registers[1] == 131


def test_LSR8_ignores_incoming_carry(vm):
    vm.flag_carry = True
    vm.registers[2] = 17000

    helper(vm, "LSR8(R1, R2)")

    assert vm.registers[1] == 66
    assert vm.flag_carry


def test_LSR8_does_not_set_carry_or_overflow(vm):
    vm.registers[2] = 15910

    helper(vm, "LSR8(R1, R2)")

    assert vm.registers[1] == 62
    assert not vm.flag_carry
    assert not vm.flag_overflow


def test_ASL_with_small_positive(vm):
    vm.registers[2] = 7

    helper(vm, "ASL(R1, R2)")

    assert vm.registers[1] == 14


def test_ASL_with_large_positive(vm):
    vm.registers[2] = 15000

    helper(vm, "ASL(R1, R2)")

    assert vm.registers[1] == 30000


def test_ASL_with_positive_overflow(vm):
    vm.registers[2] = 17000

    helper(vm, "ASL(R1, R2)")

    assert vm.registers[1] == 34000


def test_ASL_with_small_negative(vm):
    vm.registers[2] = to_u16(-7)

    helper(vm, "ASL(R1, R2)")

    assert vm.registers[1] == to_u16(-14)


def test_ASL_with_large_negative(vm):
    vm.registers[2] = to_u16(-8400)

    helper(vm, "ASL(R1, R2)")

    assert vm.registers[1] == to_u16(-16800)


def test_ASL_with_negative_overflow(vm):
    vm.registers[2] = to_u16(-20000)

    helper(vm, "ASL(R1, R2)")

    assert vm.registers[1] == 25536
    assert vm.flag_carry
    assert vm.flag_overflow


def test_ASL_shifts_out_carry_when_blocked(vm):
    vm.registers[2] = to_u16(-20000)

    helper(vm, "ASL(R1, R2)")

    assert vm.flag_carry
    assert vm.flag_overflow


def test_ASL_shifts_in_carry(vm):
    vm.flag_carry = True
    vm.registers[2] = 7

    helper(vm, "ASL(R1, R2)")

    assert vm.registers[1] == 15
    assert not vm.flag_carry


def test_ASL_ignores_carry_when_blocked(vm):
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.registers[2] = 7

    helper(vm, "ASL(R1, R2)")

    assert vm.registers[1] == 14
    assert not vm.flag_carry


def test_ASL_resets_carry(vm):
    vm.flag_carry = True

    helper(vm, "ASL(R1, R2)")

    assert not vm.flag_carry


def test_ASL_resets_overflow_flag(vm):
    vm.flag_overflow = True

    helper(vm, "ASL(R1, R2)")

    assert not vm.flag_overflow


def test_ASR_with_small_positive(vm):
    vm.registers[2] = 7

    helper(vm, "ASR(R1, R2)")

    assert vm.registers[1] == 3


def test_ASR_with_large_positive(vm):
    vm.registers[2] = 15000

    helper(vm, "ASR(R1, R2)")

    assert vm.registers[1] == 7500


def test_ASR_with_small_negative(vm):
    vm.registers[2] = to_u16(-7)

    helper(vm, "ASR(R1, R2)")

    assert vm.registers[1] == to_u16(-4)


def test_ASR_with_another_small_negative(vm):
    vm.registers[2] = to_u16(-5)

    helper(vm, "ASR(R1, R2)")

    assert vm.registers[1] == to_u16(-3)


def test_ASR_with_large_negative(vm):
    vm.registers[2] = to_u16(-8400)

    helper(vm, "ASR(R1, R2)")

    assert vm.registers[1] == to_u16(-4200)


def test_ASR_shifts_out_carry_when_blocked(vm):
    vm.flag_carry_block = True
    vm.registers[2] = 3

    helper(vm, "ASR(R1, R2)")

    assert vm.flag_carry
    assert not vm.flag_overflow


def test_ASR_ignores_incoming_carry(vm):
    vm.flag_carry = True
    vm.registers[2] = 4

    helper(vm, "ASR(R1, R2)")

    assert vm.registers[1] == 2
    assert not vm.flag_carry


def test_ASR_ignores_carry_when_blocked(vm):
    vm.flag_carry = True
    vm.flag_carry_block = True
    vm.registers[2] = 4

    helper(vm, "ASR(R1, R2)")

    assert vm.registers[1] == 2
    assert not vm.flag_carry


def test_ASR_resets_carry(vm):
    vm.flag_carry = True

    helper(vm, "ASR(R1, R2)")

    assert not vm.flag_carry


def test_ASR_does_not_reset_overflow_flag(vm):
    vm.flag_overflow = True

    helper(vm, "ASR(R1, R2)")

    assert vm.flag_overflow
